from slinn import utils


class Filter:
    def __init__(self, filter: str, methods: tuple=None):
        self.filter = filter
        self.methods = ('GET', 'POST') if methods is None else methods
        
    def check(self, text: str, method: str) -> bool:
        return utils.rematcheswith(text, self.filter) and method.upper() in self.methods
    
    def size(self, text: str, method: str) -> int:
        a = utils.min_restartswith_size(text, self.filter) if self.check(text, method) else 2147483647
        b = utils.Bmin_restartswith_size(text, self.filter) if self.check(text, method) else 2147483647
        if not self.check(text, method):
            return -1
        elif a == 2147483647:
            return 0
        else:
            return b


from slinn.http_response import HttpResponse


class HttpAPIResponse(HttpResponse):
    def __init__(self, payload: any, data: list[tuple]=None, status: str='200 OK', content_type: str='text/plain'):
        self.payload = str(payload).encode() if type(payload) in [str, int, float, bool, dict] else bytes(payload)
        self.content_type = content_type
        default_data = [('Server', 'Slinn'), ('Content-Type', content_type), ('Access-Control-Allow-Origin', '*')]
        self.data = default_data + data if data is not None else default_data + ([] if type(payload) in [str, int, float, bool, dict] else [('Content-Length', len(self.payload))])
        self.status = status



from slinn import Server, Filter


class Dispatcher:
    def __init__(self, *hosts: tuple):
        self.handles = []
        self.hosts = hosts if list(hosts) != [] else ['.*']
        
    def __call__(self, filter: Filter):
        def decorator(func):
            self.handles.append(Server.Handle(filter, func))
            return func
        return decorator	


from slinn import Request, Address, Filter, utils
import socket, ssl, time, os, traceback


RED = '\u001b[31m'
RESET = '\u001b[0m'

class Server:
    class Handle:
        def __init__(self, filter: Filter, function):
            self.filter = filter
            self.function = function
    
    def __init__(self, *dispatchers: tuple, smart_navigation: bool=True, ssl_fullchain: str=None, ssl_key: str=None, timeout: float=0.03, max_bytes_per_package: int=4096, max_bytes: int=4294967296, _func=lambda server: None):
        self.dispatchers = dispatchers
        self.smart_navigation = smart_navigation
        self.server_socket = None
        self.ssl = ssl_fullchain is not None and ssl_key is not None
        self.ssl_cert, self.ssl_key = ssl_fullchain, ssl_key
        self.ssl_context = None
        self.thread = None
        self.timeout = timeout
        self.max_bytes_per_package = max_bytes_per_package
        self.max_bytes = max_bytes
        self._func = _func

    def reload(self, *dispatchers: tuple):
        if self.thread is not None:
            self.thread.stop()
            try:
                self.thread.join()
            except RuntimeError:
                pass
        self.dispatchers = dispatchers
            

    def address(self, port: int, domain: str=None):
        protocol = 'https' if self.ssl else 'http'
        return (f'{protocol.upper()} server is available on {protocol}://'+
            ('0.0.0.0' if (domain is None or domain == '') else ('['+domain+']' if ':' in domain else domain))+
            f'{(":"+str(port) if port != 443 else "") if self.ssl else (":"+str(port )if port != 80 else "")}/')
        
    def listen(self, address: Address):		
        self.server_socket = None
        if ':' in address.host:
            self.server_socket = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
            if socket.has_dualstack_ipv6():
                self.server_socket.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, 0)
        else:
            self.server_socket = socket.socket(socket.AF_INET if '.' in address.host else socket.AF_INET6, socket.SOCK_STREAM)
        self.server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        try:
            self.server_socket.bind((address.host, address.port))
        except PermissionError:
            print(f'{RED}Permission denied{RESET}')
            exit(13)
        if self.ssl:
            self.ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
            self.ssl_context.load_cert_chain(certfile=self.ssl_cert, keyfile=self.ssl_key)
        self.server_socket.settimeout(self.timeout)
        self.server_socket.listen()
        print(self.address(address.port, address.domain))
        try:
            while True:
                try:
                    utils.StoppableThread(target=self.handle_request, args=self.server_socket.accept()).start()
                except socket.timeout:
                    pass
        except KeyboardInterrupt:
            print('Got KeyboardInterrupt, halting the application...')
            if utils.check_socket(self.server_socket):
                self.server_socket.close()
            os._exit(0)
                        
    def handle_request(self, client_socket, client_address):
        try:
            self._func(self)
            if self.ssl:
                client_socket = self.ssl_context.wrap_socket(client_socket, server_side=True)
            try:
                client_socket.settimeout(self.timeout)
                data = bytearray()
                while len(data) < self.max_bytes:
                    try:
                        b = client_socket.recv(self.max_bytes_per_package)
                        data += b
                    except TimeoutError:
                        break
                    
                data = data.split(b'\r\n\r\n')
                header = data[0].decode()
                content = b'\r\n\r\n'.join(data[1:])
                
                if header == '':
                    return

                request = Request(header, content, client_address)
                print(request)
            except KeyError:
                return print('Got KeyError, probably invalid request. Ignore')
            except UnicodeDecodeError:
                return print('Got UnicodeDecodeError, probably invalid request. Ignore')
            except ConnectionResetError:
                return print('Connection reset by client')
            except OSError:
                return print('Connection closed')
            for dispatcher in self.dispatchers:
                if True in [utils.restartswith(request.host, host) for host in dispatcher.hosts]:
                    if self.smart_navigation:
                        sizes = [handle.filter.size(request.link, request.method) for handle in dispatcher.handles]
                        if sizes != []:
                            handle = dispatcher.handles[sizes.index(max(sizes))]
                            if handle.filter.check(request.link, request.method):
                                if utils.check_socket(client_socket):
                                    data = handle.function(request).make(request.version)
                                    packages = [data[x:x+self.max_bytes_per_recieve] for x in range (0, len(data), self.max_bytes_per_recieve)]
                                    i = 0
                                    while i < len(packages):
                                        try:
                                            client_socket.sendall()
                                            i += 1
                                        except TimeoutError:
                                            continue
                                    client_socket.close()
                                return
                    else:
                        for handle in dispatcher.handles:
                            if handle.filter.check(request.link, request.method):
                                if utils.check_socket(client_socket):
                                    data = handle.function(request).make(request.version)
                                    packages = [data[x:x+self.max_bytes_per_recieve] for x in range (0, len(data), self.max_bytes_per_recieve)]
                                    client_socket.sendall()
                                    client_socket.close()
                                return
        except ssl.SSLError:
            print('During handling request, an exception has occured:')
            traceback.print_exc()
            print('Got ssl.SSLError, reloading...')
            self.reload(*self.dispatchers)
        except Exception:
            print('During handling request, an exception has occured:')
            traceback.print_exc()



import re, threading, socket


class StoppableThread(threading.Thread):
    def __init__(self,  *args: tuple, **kwargs: dict):
        super(StoppableThread, self).__init__(*args, **kwargs)
        self._stop_event = threading.Event()

    def stop(self):
        self._stop_event.set()

    def stopped(self):
        return self._stop_event.is_set()


def restartswith(text: str, reg: str):
    buf, largest = '', None
    for c in text:
        buf += c
        if re.sub(reg, '', buf) == '':
            largest = buf
    return largest is not None

def rematcheswith(text: str, reg: str): 
    return re.match('^'+reg+'$', text) is not None

def Bmin_restartswith_size(text: str, reg: str):
    buf, smallest = text, None
    for _ in range(len(text)):
        buf = buf[:-1]
        if re.sub(reg, '', buf) == '':
            smallest = buf
        else:
            break
    return len(smallest) if smallest is not None else 2147483647

def min_restartswith_size(text: str, reg: str):
    buf, smallest = text, None
    for _ in range(len(text)):
        buf = buf[:-1]
        if re.sub(reg, '', buf) == '':
            smallest = buf
    return len(smallest) if smallest is not None else 2147483647

def check_socket(sock):
    return sock.getsockopt(socket.SOL_SOCKET, socket.SO_ERROR) == 0



from slinn.http_response import HttpResponse
import json


class HttpJSONResponse(HttpResponse):
    def __init__(self, **payload: dict):
        self.data = payload['__data'] if '__data' in payload.keys() else None
        self.status = payload['__status'] if '__status' in payload.keys() else '200 OK'
        self.content_type = payload['__content_type'] if '__content_type' in payload.keys() else 'text/plain'
        if '__data' in payload: 
            del payload['__data']
        if '__status' in payload:
            del payload['__status']
        if '__content_type' in payload:
            del payload['__content_type']
        super().__init__(payload=json.dumps(payload), data=self.data, status=self.status, content_type=self.content_type)



from slinn import *


dp = Dispatcher()

@dp(LinkFilter('api'))
def api(request):
    return HttpJSONAPIResponse(status='ok')

@dp(LinkFilter(''))
@dp(LinkFilter('index'))
def index(request):
    return HttpRedirect('/helloworld')

@dp(AnyFilter)
def helloworld(request):
    return HttpResponse('Hello, world!')



migration1xx2xx = """How to migrate from 1xx(1.2.1) to 2xx(2.0.0-2.2.12)

Firstly, change all your @dp_xxx.route(...) to @dp_xxx(...)
Then, you should change some filters from Filter to AnyFilter or LinkFilter where it is suit
If you have troubles with Smart Navigation, then specify '{"smart_navigation": false}' in your project.json

If you want support of autoreloading your app, follow these steps:
- Change dp_xxx = Dispatcher(...) to dp = Dispatcher(...) and @dp_xxx.route(...) to @dp(...) OR dp = dp_xxx in the end of the file
- Execute `py manage.py migrate_app {app`s name}`
"""



import socket


class Address:
    def __init__(self, port: int, host: str=None):
        self.port = port
        if host not in [None, '']:
            family, socktype, proto, cannonname, sockaddr = socket.getaddrinfo(host, self.port, socket.AF_UNSPEC, socket.SOCK_DGRAM, 0, socket.AI_PASSIVE)[0]
            self.host = sockaddr[0]
        else:
            self.host = ''
        self.domain = host


from slinn.address import Address
from slinn.filter import Filter
from slinn.link_filter import LinkFilter
from slinn.any_filter import AnyFilter
from slinn.file import File
from slinn.request import Request
from slinn.http_response import HttpResponse
from slinn.http_redirect import HttpRedirect
from slinn.http_api_response import HttpAPIResponse
from slinn.http_json_response import HttpJSONResponse
from slinn.http_json_api_response import HttpJSONAPIResponse
from slinn.server import Server
from slinn.dispatcher import Dispatcher


version = 'Slinn Murega v2.2.14 220225A'



import urllib.parse
from slinn import File

class Request:	
    RESET = '\u001b[0m'
    GRAY = '\u001b[38;2;127;127;127m'

    @staticmethod
    def parse_http_header(http_header: str):
        result = {}
        for line in http_header.split('\r\n'):
            key, value = line.split(':')[0], ':'.join(line.split(':')[1:])
            result[key.strip()] = value.strip()
        return result
    
    @staticmethod
    def parse_http_body(http_body: bytes):
        files = [File()]
        BGNs = [b'-----------------------------', b'------WebKitFormBoundary']
        END = b'--'
        binary = 0
        for line in http_body.split(b'\r\n'):
            if binary != 2:
                if line == b'':
                    binary += 1
                elif b':' in line:
                    key, value = line.split(b':')[0].decode(), b':'.join(line.split(b':')[1:])
                    files[-1].header[key.strip()] = value.strip()
                elif files[-1].id is None:
                    if line.startswith(BGNs[0]):
                        files[-1].id = line[len(BGNs[0]):]
                    elif line.startswith(BGNs[1]):
                        files[-1].id = line[len(BGNs[1]):]
                continue
            if files[-1].id is not None:
                if line.startswith(BGNs[0]) and line.endswith(END) and line[len(BGNs[0]):-len(END)] == files[-1].id \
                   or line.startswith(BGNs[1]) and line.endswith(END) and line[len(BGNs[1]):-len(END)] == files[-1].id:
                    files[-1].data = bytes(files[-1].data[:-2])
                    files.append(File())
                    binary = 0
                    continue
            files[-1].data += line + b'\r\n'
        return {
            'payload': b'',
            'files': files[:-1]
        } if len(files[:-1]) != 0 else {
            'payload': http_body,
            'files': list()
        }

    def __init__(self, header: str, body: bytes, client_address: tuple[str, int]):
        def get_args(text):
                return {} if text == '' else {pair.split('=')[0]: '='.join(pair.split('=')[1:]) for pair in text.split('&')}
        
        self.type = header.split('\r\n')[0].strip().split(' ')
        self.header = {'method': self.type[0], 'link': ' '.join(self.type[1:-1]), 'ver': self.type[-1], 'data': {'user-agent': '', 'Accept': '', 'Accept-Encoding': '', 'Accept-Language': ''}}
        header = self.parse_http_header(header)
        self.payload = self.parse_http_body(body)['payload']
        self.files = self.parse_http_body(body)['files']
        self.header['data'].update(header)
        self.ip, self.port = client_address[:2]
        self.method = self.header['method']
        self.version = self.header["ver"]
        self.full_link = urllib.parse.unquote_plus(self.header["link"].replace('+', ' '))
        self.host = self.header["data"]["Host"]
        self.user_agent = self.header["data"]["User-Agent"] if 'User-Agent' in self.header['data'].keys() else self.header['data']['user-agent']
        self.accept = self.header["data"]["Accept"].split(',')
        self.encoding = self.header["data"]["Accept-Encoding"].split(',')
        self.language = self.header["data"]["Accept-Language"].split(',')
        self.link = self.full_link[:(self.full_link.index('?') if '?' in self.full_link else None)]
        self.args = get_args(self.full_link[(self.full_link.index('?')+1 if '?' in self.full_link else len(self.full_link)):])
        self.cookies = {c.split('=')[0]:c.split('=')[1] for c in self.header["data"]["Cookie"].split(';')} if 'Cookie' in self.header['data'] else []

    def __str__(self):
        return f'{self.GRAY}[{self.method}]{self.RESET} request {self.full_link} from {"" if "." in self.ip else "["}{self.ip}{"" if "." in self.ip else "]"}:{self.port} on {self.host}'



class HttpResponse:
    def __init__(self, payload: any, data: list[tuple]=None, status: str='200 OK', content_type: str='text/plain'):
        self.payload = str(payload).encode() if type(payload) in [str, int, float, bool] else bytes(payload)
        self.data = data if data is not None else [('Content-Type', content_type), ('Server', 'Slinn')] + ([] if type(payload) in [str, int, float, bool] else [('Content-Length', len(self.payload))])
        self.status = status
    
    def set_cookie(self, key: str, value: any):
        self.data.append(('Set-Cookie', f'{key}={value}'))
    
    def make(self, type: str='HTTP/2.0') -> str:
        return (f'{type} {self.status}'+'\r\n'+"\r\n".join([str(dat[0])+": "+str(dat[1]) for dat in self.data])+'\r\n\r\n').encode()+self.payload


import sys, os, shutil, json, base64
from slinn import version as slinn_version
import slinn

RED = '\u001b[31m'
GREEN = '\u001b[32m'
BLUE = '\u001b[34m'
RESET = '\u001b[0m'
BOLD = '\u001b[1m'
GRAY = '\u001b[38;2;127;127;127m'


def arg_parse(arg: str):
    return base64.urlsafe_b64decode(arg.removeprefix('b64@').encode()+b'==').decode() if arg.startswith('b64@') else arg

def splits(string: str, delimeters: list[str]=[' ', '\n'], quotes: list[str]=[]):
        result = ['']
        current_quote = ''
        for char in str(string):
            if char in quotes:
                if current_quote == '':
                    current_quote = char
                    continue
                elif current_quote == char:
                    current_quote = ''
                    continue
            if current_quote == '':
                if char in delimeters:
                    result.append('')
                else:
                    result[-1] += char
            else:
                result[-1] += char
        return result

def get_args(expecting, text):
    text = text.strip()
    if text == '':
        return {}
    args = {'not_used': []}
    Ds = ['\n', ' ']
    Qs = ['"', "'", '`']
    spls  = splits(text, Ds, Qs)
    i = 0
    while i < len(spls):
        arg = str(spls[i])
        try:
            if arg.strip().endswith('='):
                W = arg.strip().removesuffix('=').strip()
                expecting.pop(expecting.index(W))
                args[W] = arg_parse(str(splits[i + 1]))
                i += 2
                continue
            if len(splits(arg, ['='], Qs)) == 2:
                spl = splits(arg, ['='])
                W = str(spl[0])
                if W not in args.keys():
                    args[W] = arg_parse(str(spl[1]))
                    expecting.pop(expecting.index(W))
                else:
                    args[W] = [args[W]]
                    args[W].append(arg_parse(str(spl[1])))
                i += 1
                continue
        except ValueError:
            pass

        if len(expecting) == 0:
            args['not_used'].append(arg_parse(arg))
            i += 1
            continue

        E = expecting.pop(0)
        args[E] = arg_parse(arg)
        i += 1
    return args

def replace_all(text: str, sss: list[str] | str, ss2: str) -> str:
    for ss1 in sss:
        text = text.replace(ss1, ss2)
    return text

def add_quotes_to_list(lst: list[str]):
        for l in lst:
            yield f'\'{l}\''

def update():
    project = open('project.json', 'r')
    project_json = json.loads(project.read())
    project.close()
    if 'apps' not in project_json.keys():
        return print(f'Updated project.json')
    apps = []
    for app in project_json['apps']:
        if os.path.isdir(app):
            apps.append(app)
    project_json['apps'] = apps
    project = open('project.json', 'w')
    project.write(json.dumps(project_json, indent=4))
    project.close()
    return print(f'Updated project.json')

def main():
    if len(sys.argv) > 1:
        if sys.argv[1].lower() == 'run':
            from slinn import Server, Address

            def config():
                file = open('project.json')
                cfg = json.loads(file.read())
                file.close()
                return cfg
            
            def app_config(app):
                try:
                    cfg = {"debug": False}
                    file = open(f'{app}/config.json')
                    cfg.update(json.loads(file.read()))
                    file.close()
                    return cfg
                except FileNotFoundError:
                    print(f'{RED}{app}/config.json file not found{RESET}')
                    exit()


            def load_imports(apps, debug=False):
                imports = []
                for app in apps:
                    if not app_config(app)['debug'] or debug:
                        imports.append(f'import {app}')
                return imports


            def get_dispatchers(apps, debug=False):
                dispachers = []
                for app in apps:
                    if not app_config(app)['debug'] or debug:
                        dispachers.append(f'{app}.dp')
                return dispachers
            
            def app_reload(app):
                return f'global {app};{app} = importlib.reload({app});'
            
            print('Loading config...')
            cfg = config()
            debug = cfg["debug"] if "debug" in cfg.keys() else False
            apps = cfg['apps'] if 'apps' in cfg.keys() else []
            port = cfg['port'] if 'port' in cfg.keys() else 8080
            host = cfg['host'] if 'host' in cfg.keys() else ''
            timeout = float(cfg['timeout']) if 'timeout' in cfg.keys() else 0.03
            max_bytes_per_recieve = int(cfg['max_bytes_per_recieve']) if 'max_bytes_per_recieve' in cfg.keys() else 4096
            max_bytes = int(cfg['max_bytes']) if 'max_bytes' in cfg.keys() else 4294967296
            smart_navigation = cfg['smart_navigation'] if 'smart_navigation' in cfg.keys() else True
            ssl_fullchain, ssl_key = None, None
            if 'ssl' in cfg.keys() and 'fullchain' in cfg['ssl'].keys() and 'key' in cfg['ssl'].keys():
                ssl_fullchain = '"'+cfg['ssl']['fullchain']+'"' if cfg['ssl']['fullchain'] else None    
                ssl_key = '"'+cfg['ssl']['key']+'"' if cfg['ssl']['key'] else None
            dps = get_dispatchers(apps, debug)
            if dps == []:
                print(f'{RED}Dispatchers not found. Check your apps and ./project.json{RESET}')
                exit()
            global get_dir_checksum
            def get_dir_checksum(dir):
                import hashlib
                def get_dir_checksums(dir):
                    def md5(fname):
                        hash_md5 = hashlib.md5()
                        with open(fname, "rb") as f:
                            for chunk in iter(lambda: f.read(4096), b""):
                                hash_md5.update(chunk)
                        return hash_md5.hexdigest()
                    paths = os.listdir(dir)
                    checksums = []
                    for path in paths:
                        if os.path.isdir(path):
                            checksums += get_dir_checksums(dir+'/'+path)
                        elif path.endswith('.py'):
                            checksums.append(path+md5(dir+'/'+path))
                    return checksums
                return hashlib.md5(''.join([checksum for checksum in get_dir_checksums(dir)]).encode()).hexdigest()
            global checksum
            checksum = get_dir_checksum('.')

            reloader = """
def reloader(server):
    import importlib, traceback
    global checksum
    try:
        if checksum != get_dir_checksum('.'):
            checksum = get_dir_checksum('.')
            """
            for app in cfg['apps']:
                if not app_config(app)['debug'] or debug:
                    reloader += app_reload(app)
            reloader += """
            server.reload("""
            reloader += ",".join(dps)
            reloader += """)
    except Exception:
        print('During handling request, an exception has occured:')
        traceback.print_exc()
"""

            apps_info = []
            for app in cfg['apps']:
                if not app_config(app)['debug'] or debug:
                    apps_info.append(app)
                else:
                    apps_info.append('['+app+']')
            print(f'{GRAY}Apps: ' + ', '.join(apps_info))
            print('Debug mode ' + 'enabled' if debug else 'disabled')
            print('Smart navigation ' + ('enabled' if smart_navigation else 'disabled'))
            print(RESET)

            print('Starting server...')
            start = ';'.join(load_imports(apps, debug))+reloader+f'server=Server({",".join(dps)}, smart_navigation={smart_navigation}, ssl_fullchain={ssl_fullchain}, ssl_key={ssl_key}, timeout={timeout}, max_bytes_per_recieve={max_bytes_per_recieve}, max_bytes={max_bytes}, _func=reloader);server.listen(Address({port}, "{host}"))'
            exec(start)
        elif sys.argv[1].lower() == 'create':
            args = get_args(['name', 'host'], ' '.join(sys.argv[2:]))
            if 'name' not in args.keys():
                return print(f'{RED}The app`s name is not specified{RESET}')
            ensure_appname = replace_all(args['name'], '-&$#!@%^().,', '_')
            if os.path.isdir(ensure_appname):
                return print(f'{BLUE}The app named {args["name"]} exists{RESET}')
            if 'host' not in args.keys():
                print(f'{BLUE}Hosts were not specified')
            os.mkdir(ensure_appname)
            with open(f'{ensure_appname}/__init__.py', 'w') as f:
                data = """import sys, importlib
if '%appname%.app' not in sys.modules.keys():
    from %appname%.app import dp
else:
    dp = importlib.reload(sys.modules['%appname%.app']).dp
""".replace('%appname%', ensure_appname)
                f.write(data)
            with open(f'{ensure_appname}/app.py', 'w') as f:
                data = """from slinn import Dispatcher, Filter, HttpResponse
 
dp = Dispatcher(%hosts%)

# Write your code down here                         
""".replace('%appname%', ensure_appname).replace('%hosts%', '' if 'host' not in args.keys() else ', '.join(add_quotes_to_list(args['host'] if type(args['host']) == list else [args['host']])))
                f.write(data)
            with open(f'{ensure_appname}/config.json', 'w') as f:
                data = """
{
    "debug": false
}
"""
                f.write(data)
            fr = open('project.json', 'r')
            fj = json.loads(fr.read())
            fr.close()
            if 'apps' in fj.keys():
                fj['apps'].insert(0, ensure_appname)
            else:
                fj['apps'] = []
            fw = open('project.json', 'w')
            fw.write(json.dumps(fj, indent=4))
            fw.close()
            update()
            print(f'{GREEN}App successfully created{RESET}')
        elif sys.argv[1].lower() == 'delete':
            args = get_args(['name'], ' '.join(sys.argv[2:]))
            apppath = (args['path']+'?').replace('/?', '').replace('?', '') if 'path' in args.keys() else '.'
            if 'name' not in args.keys():
                return print(f'{RED}The app`s name is not specified{RESET}')
            ensure_appname = replace_all(args['name'], '-&$#!@%^().,', '_')
            if not os.path.isdir(ensure_appname):
                return print(f'{BLUE}The app named {args["name"]} does not exist{RESET}')
            if input(f'{RESET}Are you sure? (y/N) >>> ').lower() not in ['y', 'yes']:
                return print(f'{RESET}Aborted')
            shutil.rmtree(ensure_appname)
            if os.path.isdir(f'{apppath}/templates_data/{ensure_appname}'):
                shutil.rmtree(f'{apppath}/templates_data/{ensure_appname}')
            if len(os.listdir(f'{apppath}/templates_data')) == 0:
                shutil.rmtree(f'{apppath}/templates_data')
            update()
            return print(f'{GREEN}App successfully deleted{RESET}')
        elif sys.argv[1].lower() == 'update':
            return update()
        elif sys.argv[1].lower() == 'help':
            return print("""%BOLD%Slinn manager help page

Commands%RESET%:
    %cmd% run                                                              %GRAY%# Starts server%RESET%
    %cmd% create {app`s name} host=(host1) host=(host2)...                 %GRAY%# Creates a new app
        Example: %cmd% create localhost host=localhost host=127.0.0.1%RESET%
    %cmd% delete {app`s name} (project`s path)                             %GRAY%# Deletes an app
        Example: %cmd% delete localhost%RESET%
    %cmd% template {template`s name} (projects`s path)                     %GRAY%# Installs a template
        Example: %cmd% template firstrun%RESET%
    %cmd% update                                                           %GRAY%# Updates project.py%RESET%
    %cmd% migrate_app {app`s name}                                         %GRAY%# Migrates app(check slinn.guides.migration1xx2xx.migration1xx2xx)%RESET%
    %cmd% help                                                             %GRAY%# Prints this help%RESET%
    %cmd% version                                                          %GRAY%# Prints version of Slinn%RESET%
""".replace('%cmd%', f'py {sys.argv[0]}').replace('%GRAY%', GRAY).replace('%RESET%', RESET).replace('%BOLD%', BOLD))
        elif sys.argv[1].lower() == 'version':
            return print(slinn_version)
        elif sys.argv[1].lower() == 'template':
            args = get_args(['name', 'path'], ' '.join(sys.argv[2:]))
            apppath = (args['path']+'?').replace('/?', '').replace('?', '') if 'path' in args.keys() else '.'
            if 'name' not in args.keys():
                return print(f'{RED}Template name is not specified{RESET}')
            modulepath = os.path.abspath(slinn.__file__).replace('__init__.py', '')
            fr = open('project.json', 'r')
            fj = json.loads(fr.read())
            fr.close()
            if 'apps' in fj.keys():
                if args['name'] in fj['apps']:
                    return print(f'{BLUE}Template {args["name"]} has already installed{RESET}')
                fj['apps'].insert(0, args['name'])
            else:
                fj['apps'] = []
            fw = open('project.json', 'w')
            fw.write(json.dumps(fj, indent=4))
            fw.close()
            try:
                shutil.copytree(f'{modulepath}templates/{args["name"]}/', f'{apppath}/{args["name"]}', ignore=shutil.ignore_patterns('data'))
                try:
                    if not os.path.isdir(f'{apppath}/templates_data'):
                        os.mkdir(f'{apppath}/templates_data')
                    shutil.copytree(f'{modulepath}templates/{args["name"]}/data/', f'{apppath}/templates_data/{args["name"]}')
                except  FileExistsError:
                    pass
                except FileNotFoundError:
                    pass
                update()
                return print(f'{GREEN}Template {args["name"]} successfully installed{RESET}') 
            except FileExistsError:
                return print(f'{BLUE}Template {args["name"]} has already installed{RESET}')
            except FileNotFoundError:
                return print(f'{BLUE}Template {args["name"]} not found{RESET}')
        elif sys.argv[1].lower() == 'migrate_app':
            args = get_args(['name'], ' '.join(sys.argv[2:]))
            if 'name' not in args.keys():
                return print(f'{RED}The app`s name is not specified{RESET}')
            ensure_appname = replace_all(args['name'], '-&$#!@%^().,', '_')
            if not os.path.isdir(ensure_appname):
                return print(f'{BLUE}The app named {args["name"]} does not exist{RESET}')
            with open(f'{ensure_appname}/__init__.py', 'w') as f:
                data = """import sys, importlib
if '%appname%.app' not in sys.modules.keys():
    from %appname%.app import dp
else:
    dp = importlib.reload(sys.modules['%appname%.app']).dp
""".replace('%appname%', ensure_appname)
                f.write(data)
            with open(f'{ensure_appname}/config.json', 'w') as f:
                data = """
{
    "debug": false
}
"""
                f.write(data)
            print(f'{GREEN}App successfully migrated{RESET}')
        else:
            return print(f'{RED}Command {sys.argv[1].lower()} is not exists{RESET}')
    else:
        return print(f'{RED}Command was not specified{RESET}')

            
if __name__ == '__main__':
    main()


import sys, venv
from slinn.default import *


RED = '\u001b[31m'
GREEN = '\u001b[32m'
BLUE = '\u001b[34m'
RESET = '\u001b[0m'
BOLD = '\u001b[1m'
GRAY = '\u001b[38;2;127;127;127m'


if __name__ == '__main__':
    if len(sys.argv) > 1:
        if sys.argv[1].lower() == 'create':
            args = get_args(['path'], ' '.join(sys.argv[2:]))
            apppath = (args['path']+'?').replace('/?', '').replace('?', '') if 'path' in args.keys() else '.'
            import os, slinn, shutil, platform
            modulepath = os.path.abspath(slinn.__file__).replace('__init__.py', '')
            if not os.path.isdir(apppath):
                os.mkdir(apppath)
            else:
                print(f'{BLUE}{apppath} has already existed{RESET}')
            shutil.copyfile(modulepath+'default/manage.py', f'{apppath}/manage.py')
            shutil.copyfile(modulepath+'default/project.json', f'{apppath}/project.json')
            venv.create(f'{apppath}/venv')
            packages_dir = ''
            if platform.system() == 'Windows':
                packages_dir = f'{apppath}/venv/Lib/site-packages'
            else:
                packages_dir = f'{apppath}/venv/lib/python{".".join(sys.version.split(" ")[0].split(".")[:-1])}/site-packages'
            try:
                os.makedirs(packages_dir)
            except FileExistsError:
                pass
            try:
                shutil.copytree(modulepath, packages_dir+'/slinn')
            except FileNotFoundError:
                print(f'{RED}Cannot install slinn to the new virtual environment{RESET}')
                exit()
            try:
                import pip
                shutil.copytree(os.path.abspath(pip.__file__).replace('__init__.py', ''), packages_dir+'/pip')
            except FileNotFoundError:
                print(f'{BLUE}pip was not installed{RESET}')
            try:
                import wheel
                shutil.copytree(os.path.abspath(wheel.__file__).replace('__init__.py', ''), packages_dir+'/wheel')
            except Exception:
                print(f'{BLUE}wheel was not installed{RESET}')
            try:
                import setuptools
                shutil.copytree(os.path.abspath(setuptools.__file__).replace('__init__.py', ''), packages_dir+'/setuptools')
            except Exception:
                print(f'{BLUE}setuptools was not installed{RESET}')
            print(f'{GREEN}Project has created{RESET}')
            modulepath = os.path.abspath(slinn.__file__).replace('__init__.py', '')
            try:
                shutil.copytree(f'{modulepath}templates/firstrun/', f'{apppath}/firstrun')
                shutil.copytree(f'{modulepath}templates/firstrun/data/', f'{apppath}/templates_data/firstrun')
                print(f'{GREEN}Template firstrun successfully installed{RESET}') 
            except FileExistsError:
                print(f'{BLUE}Template firstrun has already existed installed{RESET}')
            except FileNotFoundError:
                print(f'{BLUE}Template firstrun not found{RESET}')
        elif sys.argv[1].lower() == 'update':
            args = get_args(['path'], ' '.join(sys.argv[2:]))
            apppath = (args['path']+'?').replace('/?', '').replace('?', '') if 'path' in args.keys() else '.'
            import os, slinn, shutil, platform
            modulepath = os.path.abspath(slinn.__file__).replace('__init__.py', '')
            if not os.path.isdir(apppath):
                print(f'{BLUE}`{apppath}` does not exist{RESET}')
                exit()
            packages_dir = ''
            if platform.system() == 'Windows':
                packages_dir = f'{apppath}/venv/Lib/site-packages'
            else:
                packages_dir = f'{apppath}/venv/lib/python{".".join(sys.version.split(" ")[0].split(".")[:-1])}/site-packages'
            if not os.path.isdir(packages_dir+'/slinn'):
                print(f'{RED}Virtual environment directory is corrupted. Reinstall the project{RESET}')
                exit()
            if not os.path.isfile(apppath+'/manage.py'):
                print(f'{RED}manage.py file does not exist. Reinstall the project{RESET}')
                exit()
            shutil.rmtree(packages_dir+'/slinn')
            os.remove(apppath+'/manage.py')
            shutil.copytree(modulepath, packages_dir+'/slinn')
            shutil.copyfile(modulepath+'default/manage.py', f'{apppath}/manage.py')
            print(f'{GREEN}Project has updated{RESET}')
        elif sys.argv[1].lower() == 'help':
            print("""%BOLD%Slinn help page

Commands%RESET%:
    %cmd% create {project`s name}		%GRAY%# Creates a new project%RESET%
    %cmd% update {project`s name}		%GRAY%# Updates the project%RESET%
    %cmd% help                   		%GRAY%# Prints this help%RESET%
    %cmd% version                		%GRAY%# Prints version of Slinn%RESET%
""".replace('%cmd%', f'py -m slinn').replace('%GRAY%', GRAY).replace('%RESET%', RESET).replace('%BOLD%', BOLD))
        elif sys.argv[1].lower() == 'version':
            print(slinn_version)
        else:
            print(f'{RED}Command {sys.argv[1].lower()} is not exists{RESET}')
    else:
        print(f'{RED}Command was not specified{RESET}')




from slinn.filter import Filter


class LinkFilter(Filter):
    def __init__(self, _filter: str, methods: list[str]=None):
        self.filter = r'\/?' + _filter.replace('/', r'\/') + r'(\/.*)?'
        self.methods = ['GET', 'POST'] if methods is None else methods



from slinn.http_json_response import HttpJSONResponse


class HttpJSONAPIResponse(HttpJSONResponse):
    def __init__(self, **payload: dict):
        super().__init__(**payload)
        default_data = [('Server', 'Slinn'), ('Content-Type', self.content_type), ('Access-Control-Allow-Origin', '*')]
        self.data = default_data + self.data if self.data is not None else default_data



from setuptools import setup

def readme():
    with open('README.md', 'r') as f:
        return f.read()
    
templates = ['firstrun', 'example']


setup(name='slinn',
      version='2.2.14',
      description='A HTTPS and HTTP server framework',
      packages=['slinn', 'slinn.templates', 'slinn.guides'] + ['slinn.templates.' + template for template in templates],
      package_data={'slinn': ['default/*.*']} | {'slinn.templates.' + template: ['data/*.css', 'data/*.html', 'config.json'] for template in templates},
      author='Mark Radin',
      author_email='mrybs2@gmail.com',
      url='https://slinn.miotp.ru/',
      long_description=readme(),
      long_description_content_type='text/markdown',
      python_requires='>=3.11',
      zip_safe=True)



from slinn import Dispatcher, LinkFilter, AnyFilter, HttpResponse
import slinn

def read(filename):
    file = open(filename, 'r', encoding='utf-8')
    data = file.read()
    file.close()
    return data

dp = Dispatcher()

@dp(LinkFilter('styles.css'))
def styles(request):
    return HttpResponse(read('templates_data/firstrun/styles.css'))

@dp(AnyFilter)
def index(request):
    return HttpResponse(read('templates_data/firstrun/slinn.html').replace('{% version %}', slinn.version.split()[2]), content_type='text/html')
